@@ -131,6 +131,39 @@ class SendingDataFinishCommand:
         d.write_int(self.command_length)
         return d.to_bytes()
 
+class GetDisplayInfoCommand:
+    """
+    Allows retrieving display parameters.
+    """
+    def serialize(self):
+        d = ByteWriter()
+        d.write_byte(4) # length
+        d.write_byte(18) # GetDisplayInfoCommand
+        d.write_short(0)
+        return d.to_bytes()
+
+class GetVersionCommand:
+    """
+    Allows retrieving device version info.
+    """
+    def serialize(self):
+        d = ByteWriter()
+        d.write_byte(4) # length
+        d.write_byte(16) # GetVersionCommand
+        d.write_short(0)
+        return d.to_bytes()
+
+class GetBufferSizeCommand:
+    """
+    Allows retrieving data buffer size.
+    """
+    def serialize(self):
+        d = ByteWriter()
+        d.write_byte(4) # length
+        d.write_byte(20) # GetBufferSizeCommand
+        d.write_short(0)
+        return d.to_bytes()
+
 class SendDataCommand:
     """
     The main command used to send data to the device
@@ -492,6 +525,46 @@ class ContinueSendingResponse:
         self.command_type = d.read_short()
         self.continue_from = d.read_int()
 
+class DisplayInfoResponse:
+    """
+    Contains response from GetDisplayInfoCommand
+    """
+    def __init__(self, content):
+        assert len(content) == 11
+        d = ByteReader(content)
+        d.read_bytes(2) # junk data?
+        assert d.read_byte() == 0
+        self.width = d.read_short()
+        self.height = d.read_short()
+        self.color_depth = d.read_byte()
+        self.frame_limit = d.read_byte()
+        self.brightness = d.read_byte()
+        self.font_info = d.read_byte()
+
+class VersionResponse:
+    """
+    Contains response from GetVersionCommand
+    """
+    def __init__(self, content):
+        assert len(content) == 13
+        d = ByteReader(content)
+        d.read_bytes(2) # junk data?
+        assert d.read_byte() == 0
+        self.device_type = d.read_short()
+        self.device_revision = d.read_int()
+        self.software_revision = d.read_int()
+
+class BufferSizeResponse:
+    """
+    Contains response from GetBufferSizeCommand
+    """
+    def __init__(self, content):
+        assert len(content) == 7
+        d = ByteReader(content)
+        d.read_bytes(2) # junk data?
+        assert d.read_byte() == 0
+        self.buffer_size = d.read_int()
+
 def getCommandResponse(data):
     response = GenericCommandResponse(data)
 
@@ -501,6 +574,15 @@ def getCommandResponse(data):
     if (response.command_type == 255):
         return ContinueSendingResponse(response.content)
 
+    if (response.command_type == 19):
+        return DisplayInfoResponse(response.content)
+
+    if (response.command_type == 17):
+        return VersionResponse(response.content)
+
+    if (response.command_type == 21):
+        return BufferSizeResponse(response.content)
+
     return response
 
 def parse_yaff_font(fontfile):
@@ -695,6 +777,13 @@ class LedConnection:
         self.data_serial_no = 0
         self.command_serial_no = 0
 
+        self.buffer_size = self.query_command(GetBufferSizeCommand()).buffer_size
+        display_info = self.query_command(GetDisplayInfoCommand())
+        self.width = display_info.width
+        self.height = display_info.height
+        self.frame_limit = display_info.frame_limit
+        self.brightness = display_info.brightness
+
     def _on_notification(self, handle, data):
         if handle == self.cmd_handle:
             self.last_data = data
@@ -730,28 +819,39 @@ class LedConnection:
         self._ensure_connection()
         self.current_wait_event.clear()
         self.connection.write_cmd(self.cmd_handle, command.serialize())
-    
-    def wait_for_response(self):
+
+    def query_command(self, command, timeout=0.2, attempts=5):
+        """
+        Send a control command to the device and wait for a response.
+        Used for basic commands and data sending flow control.
+        """
+        for i in range(attempts + 1):
+            try:
+                self._ensure_connection()
+                self.current_wait_event.clear()
+                self.connection.write_cmd(self.cmd_handle, command.serialize())
+                return self.wait_for_response(timeout)
+            except TimeoutError:
+                if i == attempts:
+                    raise
+                self.connection.disconnect()
+
+    def wait_for_response(self, timeout=0.2):
         """
         Wait for and return a response, usually from a command sent via send_command.
         """
-        if not self.current_wait_event.wait(5):
+        if not self.current_wait_event.wait(timeout):
             raise TimeoutError("Timeout exceeded waiting for GATT response.")
         return getCommandResponse(self.last_data)
 
-    def send_data(self, data_command):
-        """
-        Send a data command to the device.
-        Currently only SendDataCommand is used, which accepts raw serialized data.
-        """
+    def _send_data_internal(self, data_command, timeout=0.2):
         self._ensure_connection()
         data_command.serial_no = self._next_data_serial_no()
         serial_no = self._next_command_serial_no()
 
         payload = data_command.serialize()
         self.send_command(SendingDataStartCommand(serial_no, data_command.command_type, len(payload)))
-        
-        response = self.wait_for_response()
+        response = self.wait_for_response(timeout)
         assert type(response) == SendingDataResponse
         assert response.serial_no == serial_no
         assert response.command_type == data_command.command_type
@@ -760,7 +860,7 @@ class LedConnection:
         seek = 0
         sent_payloads = 0
         send_size = 20
-        send_count = 6
+        send_count = self.buffer_size // send_size
 
         while seek < len(payload):
             self.current_wait_event.clear()
@@ -770,20 +870,35 @@ class LedConnection:
 
             if sent_payloads >= send_count:
                 sent_payloads = 0
-                response = self.wait_for_response()
+                response = self.wait_for_response(timeout)
                 assert type(response) == ContinueSendingResponse
                 assert response.serial_no == serial_no
                 assert response.command_type == data_command.command_type
                 seek = response.continue_from
         
         self.send_command(SendingDataFinishCommand(serial_no, data_command.command_type, len(payload)))
-        self.wait_for_response()
+        self.wait_for_response(timeout)
+
+    def send_data(self, data_command, timeout=0.2, attempts=5):
+        """
+        Send a data command to the device.
+        Currently only SendDataCommand is used, which accepts raw serialized data.
+        """
+        for i in range(attempts + 1):
+            try:
+                self._send_data_internal(data_command, timeout)
+                return
+            except TimeoutError:
+                if i == attempts:
+                    raise
+                self.connection.disconnect()
 
     def set_brightness(self, brightness):
         """
         Sets the display brightness. 0 is lowest and 100 is highest.
         """
         self.send_data(SendDataCommand(BrightnessData(brightness).serialize()))
+        self.brightness = brightness
 
     def set_screen_mode(self, mode: ScreenMode):
         """
@@ -791,7 +906,7 @@ class LedConnection:
         """
         self.send_data(SendDataCommand(ScreenModeData(mode.value).serialize()))
 
-    def set_text_by_chars(self, text, effect=Effect.SCROLL_LEFT, font="6x12", speed=0, min_height=12, char_limit=72):
+    def set_text_by_chars(self, text, effect=Effect.SCROLL_LEFT, font="6x12", speed=0, char_limit=72):
         """
         Sends text as characters. The device decides how to display them.
         This tends to be slower and more limited than set_text which sends the text as an animation.
@@ -800,33 +915,31 @@ class LedConnection:
             raise ValueError("The text exceeds the device character limit.")
 
         font_data = find_and_load_font(font)
-        font_characters = create_font_characters(text, font_data, min_height)
+        font_characters = create_font_characters(text, font_data, self.height)
         font_character_data = SendDataCommand(FontData(font_characters).serialize())
         text_data = SendDataCommand(TextData(text, speed, effect).serialize())
         self.send_data(font_character_data)
         self.send_data(text_data)
 
-    def set_text_lines(self, text, align=Align.CENTER, font="4x6", frame_duration=2, width=48,
-            lines_per_frame=2, line_height=6, effect=Effect.NONE, speed=20, reflow=True,
-            frame_limit=20):
+    def set_text_lines(self, text, align=Align.CENTER, font="4x6", frame_duration=2, line_height=6,
+            effect=Effect.NONE, speed=20, reflow=True):
         """
         Sends multi-line text as an animation. Can pack two lines of text onto the display.
         """
         font_data = find_and_load_font(font)
 
         if reflow:
-            lines = reflow_text(text, font_data, width)
+            lines = reflow_text(text, font_data, self.width)
         else:
             lines = text.replace('\r', '').split('\n')
 
-        frames = lines_to_frames(lines, font_data, align, width, lines_per_frame, line_height)
-        if len(frames) > frame_limit:
+        frames = lines_to_frames(lines, font_data, align, self.width, self.height // line_height, line_height)
+        if len(frames) > self.frame_limit:
             raise ValueError("The animation exceeds the device frame limit.")
 
-        height = lines_per_frame * line_height
         frame_data = SendDataCommand(
             AnimationData(
-                [FrameData(width, height, gen_bitmap(*frame)) for frame in frames],
+                [FrameData(self.width, self.height, gen_bitmap(*frame)) for frame in frames],
                 int(frame_duration * 1000),
                 speed,
                 effect
@@ -835,7 +948,7 @@ class LedConnection:
 
         self.send_data(frame_data)
 
-    def set_text(self, text, effect=Effect.SCROLL_LEFT, font="6x12", speed=0, min_height=12):
+    def set_text(self, text, effect=Effect.SCROLL_LEFT, font="6x12", speed=0):
         """
         Sends single-line scrolling text as an animation.
         """
@@ -843,20 +956,19 @@ class LedConnection:
             text,
             Align.LEFT,
             font,
-            lines_per_frame=1,
-            line_height=min_height,
+            line_height=self.height,
             effect=effect,
             speed=speed,
             reflow=False
         )
 
-    def clear(self, width=48, height=12):
+    def clear(self):
         """
         Clears the display by sending an empty frame.
         """
         frame_data = SendDataCommand(
             AnimationData(
-                [FrameData(width, height, b'\x00' * int(width * height / 8))],
+                [FrameData(self.width, self.height, b'\x00' * int(self.width * self.height / 8))],
                 0,
                 0,
                 Effect.NONE
